"""Provides RequestProcessor class for processing a single request."""
import inspect
import logging
from typing import (
    Any,
    Callable,
    Optional,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from jsonrpcobjects.errors import INTERNAL_ERROR
from jsonrpcobjects.objects import (
    ErrorObjectData,
    ErrorResponseObject,
    NotificationObject,
    NotificationObjectParams,
    NotificationType,
    RequestObject,
    RequestType,
    ResultResponseObject,
)

__all__ = ("RequestProcessor",)

log = logging.getLogger("openrpc")


class RequestProcessor:
    """Execute a JSON-RPC2 request and get the response."""

    def __init__(
        self,
        method: Callable,
        uncaught_error_code: Optional[int],
        request: Union[RequestType, NotificationType],
    ) -> None:
        self.method = method
        self.request = request
        self.uncaught_error_code = uncaught_error_code

    def execute(self) -> Optional[str]:
        """Execute the method and get the JSON-RPC2 response."""
        try:
            result = self._execute()
            if isinstance(self.request, (NotificationObject, NotificationObjectParams)):
                # If request was notification, return nothing.
                return None
            return ResultResponseObject(id=self.request.id, result=result).json()

        except Exception as e:
            return self._get_error_response(e)

    async def execute_async(self) -> Any:
        """Execute the method and get the JSON-RPC2 response.

        If the method is an async method it will be awaited.
        """
        try:
            result = self._execute()
            if inspect.isawaitable(result):
                result = await result
            if isinstance(self.request, (NotificationObject, NotificationObjectParams)):
                # If request was notification, return nothing.
                return None
            return ResultResponseObject(id=self.request.id, result=result).json()

        except Exception as e:
            return self._get_error_response(e)

    def _execute(self) -> Any:
        annotations = get_type_hints(self.method)

        # Call method.
        if isinstance(self.request, (RequestObject, NotificationObject)):
            result = self.method()
        elif isinstance(self.request.params, list):
            result = self.method(
                *self._get_list_params(self.request.params, annotations)
            )
        elif isinstance(self.request.params, dict):
            result = self.method(
                **self._get_dict_params(self.request.params, annotations)
            )
        else:
            result = self.method()
        return result

    def _get_error_response(self, e: Exception):
        log.exception(f"{type(e).__name__}:")
        if self.uncaught_error_code:
            return ErrorResponseObject(
                id=self.request.id,
                error=ErrorObjectData(
                    code=self.uncaught_error_code,
                    message="Server error",
                    data=f"{type(e).__name__}: {e}",
                ),
            ).json()
        return ErrorResponseObject(id=self.request.id, error=INTERNAL_ERROR).json()

    def _get_list_params(self, params: list, annotations: dict) -> list:
        try:
            return [
                self._deserialize(p, list(annotations.values())[i])
                for i, p in enumerate(params)
            ]
        except IndexError:
            return params

    def _get_dict_params(self, params: dict, annotations: dict) -> dict:
        try:
            return {k: self._deserialize(p, annotations[k]) for k, p in params.items()}
        except KeyError:
            return params

    def _deserialize(self, param: Any, p_type: Type) -> Any:
        """Deserialize dict to python objects."""
        if get_origin(p_type) == Union:
            for arg in get_args(p_type):
                try:
                    return self._deserialize(param, arg)
                except TypeError:
                    continue
        if get_origin(p_type) == list:
            types = get_args(p_type)
            return [self._deserialize(it, types[0]) for it in param]
        if not isinstance(param, dict):
            return param
        try:
            return p_type(**param)
        except (TypeError, AttributeError, KeyError):
            return param



"""Provide JSON-RPC2 server class."""
import asyncio
import json
import logging
from dataclasses import dataclass
from json import JSONDecodeError
from typing import Any, Callable, Optional, Type, Union

from jsonrpcobjects.errors import INVALID_REQUEST, METHOD_NOT_FOUND, PARSE_ERROR
from jsonrpcobjects.objects import (
    ErrorObjectData,
    ErrorResponseObject,
    NotificationObject,
    NotificationObjectParams,
    NotificationType,
    RequestObject,
    RequestObjectParams,
    RequestType,
)
from pydantic import ValidationError

from openrpc._request_processor import RequestProcessor
from openrpc.objects import MethodObject

__all__ = ("RPCServer",)

T = Type[Callable]
log = logging.getLogger("openrpc")
NotificationTypes = (NotificationObject, NotificationObjectParams)
RequestTypes = (RequestObject, RequestObjectParams)


@dataclass
class _RegisteredMethod:
    fun: Callable
    method: MethodObject


class RPCServer:
    """Class to register and methods execute methods."""

    def __init__(self, server_error_code: int) -> None:
        self.methods: dict[str, _RegisteredMethod] = {}
        self.uncaught_error_code: Optional[int] = server_error_code

    def method(self, func: T, method: MethodObject) -> T:
        """Register a method with this server for later calls."""
        log.debug("Registering method [%s]", func.__name__)
        method.name = method.name or func.__name__
        self.methods[method.name] = _RegisteredMethod(func, method)
        return func

    def process(self, data: Union[bytes, str]) -> Optional[str]:
        """Process a JSON-RPC2 request and get the response.

        :param data: A JSON-RPC2 request.
        :return: A valid JSON-RPC2 response.
        """
        parsed_json = _get_parsed_json(data)
        if isinstance(parsed_json, ErrorResponseObject):
            return parsed_json.json()

        # Batch
        if isinstance(parsed_json, list):
            requests = [_get_request_object(it) for it in parsed_json]
            results = []
            for it in requests:
                if it.method not in self.methods.keys():
                    results.append(_get_method_not_found_error(it))

                fun = self.methods[it.method].fun
                if isinstance(it, ErrorResponseObject):
                    results.append(it.json())
                elif isinstance(it, RequestTypes):
                    results.append(
                        RequestProcessor(fun, self.uncaught_error_code, it).execute()
                    )
                else:
                    # To get here, it must be a notification.
                    RequestProcessor(fun, self.uncaught_error_code, it).execute()
            return f"[{','.join(results)}]"

        # Single Request
        req = _get_request_object(parsed_json)
        if isinstance(req, ErrorResponseObject):
            return req.json()
        if req.method not in self.methods.keys():
            return _get_method_not_found_error(req)
        result = RequestProcessor(
            self.methods[req.method].fun, self.uncaught_error_code, req
        ).execute()
        return None if isinstance(req, NotificationTypes) else result

    async def process_async(self, data: Union[bytes, str]) -> Optional[str]:
        """Process a JSON-RPC2 request and get the response.

        If the method called by the request is async it will be awaited.

        :param data: A JSON-RPC2 request.
        :return: A valid JSON-RPC2 response.
        """
        parsed_json = _get_parsed_json(data)
        if isinstance(parsed_json, ErrorResponseObject):
            return parsed_json.json()

        # Batch
        if isinstance(parsed_json, list):

            async def _one_iter(it) -> Any:
                if it.method not in self.methods.keys():
                    return _get_method_not_found_error(it)

                fun = self.methods[it.method].fun
                if isinstance(it, ErrorResponseObject):
                    return it.json()
                elif isinstance(it, RequestTypes):
                    return await RequestProcessor(
                        fun, self.uncaught_error_code, it
                    ).execute_async()
                # To get here, it must be a notification.
                await RequestProcessor(
                    fun, self.uncaught_error_code, it
                ).execute_async()

            results = await asyncio.gather(
                *[_one_iter(_get_request_object(it)) for it in parsed_json]
            )
            return f"[{','.join(str(r) for r in results if r is not None)}]"

        # Single Request
        req = _get_request_object(parsed_json)
        if isinstance(req, ErrorResponseObject):
            return req.json()
        if req.method not in self.methods.keys():
            return _get_method_not_found_error(req)
        result = await RequestProcessor(
            self.methods[req.method].fun, self.uncaught_error_code, req
        ).execute_async()
        return None if isinstance(req, NotificationTypes) else result


def _get_method_not_found_error(req: Union[NotificationType, RequestType]) -> str:
    return ErrorResponseObject(
        id=None if isinstance(req, NotificationTypes) else req.id,
        error=ErrorObjectData(**{**METHOD_NOT_FOUND.dict(), **{"data": req.method}}),
    ).json()


def _get_parsed_json(data: Union[bytes, str]) -> Union[ErrorResponseObject, dict, list]:
    try:
        parsed_json = json.loads(data)
    except (TypeError, JSONDecodeError) as e:
        log.exception(f"{type(e).__name__}:")
        return ErrorResponseObject(error=PARSE_ERROR)

    if not isinstance(parsed_json, (list, dict)):
        log.error("Invalid request [%s]", parsed_json)
        return ErrorResponseObject(
            error=ErrorObjectData(**{**INVALID_REQUEST.dict(), **{"data": parsed_json}})
        )
    return parsed_json


def _get_request_object(
    data: dict[str, Any]
) -> Union[ErrorResponseObject, NotificationType, RequestType]:
    is_request = data.get("id") is not None
    has_params = data.get("params") is not None

    try:
        if is_request:
            return (RequestObjectParams if has_params else RequestObject)(**data)
        return (NotificationObjectParams if has_params else NotificationObject)(**data)
    except (TypeError, ValidationError) as e:
        log.exception(f"{type(e).__name__}:")
        return ErrorResponseObject(
            id=data.get("id"),
            error=ErrorObjectData(**{**INVALID_REQUEST.dict(), **{"data": data}}),
        )



"""Python class representations of OpenRPC and JSON Schema objects."""
from __future__ import annotations

from enum import Enum
from typing import Any, Optional, Union

from pydantic import BaseModel, Field


class ParamStructure(Enum):
    """OpenRPC ParamStructure options."""

    BY_NAME = "by-name"
    BY_POSITION = "by-position"
    EITHER = "either"


class SchemaObject(BaseModel):
    """JSON Schema object."""

    id: Optional[str] = Field(alias="$id", default=None)
    title: Optional[str] = None
    format: Optional[str] = None
    type: Optional[Union[str, list[str]]] = None
    all_of: Optional[list[SchemaObject]] = Field(alias="allOf", default=None)
    any_of: Optional[list[SchemaObject]] = Field(alias="anyOf", default=None)
    one_of: Optional[list[SchemaObject]] = Field(alias="oneOf", default=None)
    not_: Optional[SchemaObject] = Field(alias="not", default=None)
    pattern: Optional[str] = None
    minimum: Optional[float] = None
    maximum: Optional[float] = None
    exclusive_minimum: Optional[float] = Field(alias="exclusiveMinimum", default=None)
    exclusive_maximum: Optional[float] = Field(alias="exclusiveMaximum", default=None)
    multiple_of: Optional[float] = Field(alias="multipleOf", default=None)
    min_length: Optional[int] = Field(alias="minLength", default=None)
    max_length: Optional[int] = Field(alias="maxLength", default=None)
    properties: Optional[dict[str, SchemaObject]] = None
    pattern_properties: Optional[dict[str, SchemaObject]] = Field(
        alias="patternProperties", default=None
    )
    additional_properties: Optional[Union[bool, dict[str, Any]]] = Field(
        alias="additionalProperties", default=None
    )
    property_names: Optional[SchemaObject] = Field(alias="propertyNames", default=None)
    min_properties: Optional[int] = Field(alias="minProperties", default=None)
    max_properties: Optional[int] = Field(alias="maxProperties", default=None)
    required: Optional[list[str]] = None
    definitions: Optional[dict[str, SchemaObject]] = None
    items: Optional[Union[SchemaObject, bool]] = None
    prefix_items: Optional[list[SchemaObject]] = Field(
        alias="prefixItems", default=None
    )
    contains: Optional[SchemaObject] = None
    min_contains: Optional[int] = Field(alias="minContains", default=None)
    max_contains: Optional[int] = Field(alias="maxContains", default=None)
    min_items: Optional[int] = Field(alias="minItems", default=None)
    max_items: Optional[int] = Field(alias="maxItems", default=None)
    unique_items: Optional[bool] = Field(alias="uniqueItems", default=None)
    ref: Optional[str] = Field(alias="$ref", default=None)
    description: Optional[str] = None
    deprecated: Optional[bool] = None
    default: Optional[Any] = None
    examples: Optional[list[Any]] = None
    read_only: Optional[bool] = Field(alias="readOnly", default=None)
    write_only: Optional[bool] = Field(alias="writeOnly", default=None)
    const: Optional[Any] = None
    dependent_required: Optional[dict[str, list[str]]] = Field(
        alias="dependentRequired", default=None
    )
    dependent_schemas: Optional[dict[str, SchemaObject]] = Field(
        alias="dependentSchemas", default=None
    )
    if_: Optional[SchemaObject] = Field(alias="if", default=None)
    then: Optional[SchemaObject] = None
    else_: Optional[SchemaObject] = Field(alias="else", default=None)
    schema_dialect: Optional[str] = Field(alias="$schema", default=None)


# noinspection PyMissingOrEmptyDocstring
class ServerVariableObject(BaseModel):
    default: str
    enum: Optional[list[str]] = None
    description: Optional[str] = None


# noinspection PyMissingOrEmptyDocstring
class ServerObject(BaseModel):
    name: str
    url: str = "localhost"
    summary: Optional[str] = None
    description: Optional[str] = None
    variables: Optional[dict[str, ServerVariableObject]] = None


# noinspection PyMissingOrEmptyDocstring
class ContactObject(BaseModel):
    name: Optional[str] = None
    url: Optional[str] = None
    email: Optional[str] = None


# noinspection PyMissingOrEmptyDocstring
class LicenseObject(BaseModel):
    name: str
    url: Optional[str] = None


# noinspection PyMissingOrEmptyDocstring
class InfoObject(BaseModel):
    title: str
    version: str
    description: Optional[str] = None
    termsOfService: Optional[str] = None
    contact: Optional[ContactObject] = None
    license: Optional[LicenseObject] = None


# noinspection PyMissingOrEmptyDocstring
class ContentDescriptorObject(BaseModel):
    name: str
    json_schema: SchemaObject = Field(alias="schema")
    summary: Optional[str] = None
    description: Optional[str] = None
    required: Optional[bool] = None
    deprecated: bool = False


# noinspection PyMissingOrEmptyDocstring
class ExternalDocumentationObject(BaseModel):
    url: str
    description: Optional[str] = None


# noinspection PyMissingOrEmptyDocstring
class TagObject(BaseModel):
    name: str
    summary: Optional[str] = None
    description: Optional[str] = None
    externalDocs: Optional[ExternalDocumentationObject] = None


# noinspection PyMissingOrEmptyDocstring
class ErrorObject(BaseModel):
    code: int
    message: str
    data: Any = None


# noinspection PyMissingOrEmptyDocstring
class LinkObject(BaseModel):
    name: str
    description: Optional[str] = None
    summary: Optional[str] = None
    method: Optional[str] = None
    params: Optional[dict[str, Any]] = None
    server: Optional[ServerObject] = None


# noinspection PyMissingOrEmptyDocstring
class ExampleObject(BaseModel):
    name: Optional[str] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    value: Any = None
    externalValue: Optional[str] = None


# noinspection PyMissingOrEmptyDocstring
class ExamplePairingObject(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    summary: Optional[str] = None
    params: Optional[ExampleObject] = None
    result: Optional[ExampleObject] = None


# noinspection PyMissingOrEmptyDocstring
class MethodObject(BaseModel):
    name: Optional[str] = None
    params: Optional[list[ContentDescriptorObject]] = None
    result: Optional[ContentDescriptorObject] = None
    tags: Optional[list[TagObject]] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    externalDocs: Optional[ExternalDocumentationObject] = None
    deprecated: Optional[bool] = None
    servers: Optional[list[ServerObject]] = None
    errors: Optional[list[ErrorObject]] = None
    links: Optional[list[LinkObject]] = None
    paramStructure: Optional[ParamStructure] = None
    examples: Optional[list[ExamplePairingObject]] = None


# noinspection PyMissingOrEmptyDocstring
class ComponentsObject(BaseModel):
    contentDescriptors: Optional[dict[str, ContentDescriptorObject]] = None
    schemas: Optional[dict[str, SchemaObject]] = None
    examples: Optional[dict[str, ExampleObject]] = None
    links: Optional[dict[str, LinkObject]] = None
    errors: Optional[dict[str, ErrorObject]] = None
    examplePairingObjects: Optional[dict[str, ExamplePairingObject]] = None
    tags: Optional[dict[str, TagObject]] = None


# noinspection PyMissingOrEmptyDocstring
class OpenRPCObject(BaseModel):
    openrpc: str
    info: InfoObject
    methods: list[MethodObject]
    servers: Optional[Union[ServerObject, list[ServerObject]]] = None
    components: Optional[ComponentsObject] = None
    externalDocs: Optional[ExternalDocumentationObject] = None


SchemaObject.update_forward_refs()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['openrpc']

package_data = \
{'': ['*']}

install_requires = \
['jsonrpc2-objects>=1.3.7,<2.0.0', 'pydantic>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'openrpc',
    'version': '1.2.10',
    'description': 'OpenRPC provides classes to rapidly develop an OpenRPC server.',
    'long_description': '<div align=center>\n  <h1>OpenRPC</h1>\n  <img src="https://img.shields.io/badge/License-AGPL%20v3-blue.svg"\n   height="20"\n   alt="License: AGPL v3">\n  <img src="https://img.shields.io/badge/code%20style-black-000000.svg"\n   height="20"\n   alt="Code style: black">\n  <img src="https://img.shields.io/pypi/v/openrpc.svg"\n   height="20"\n   alt="PyPI version">\n  <a href="https://gitlab.com/mburkard/openrpc/-/blob/main/CONTRIBUTING.md">\n    <img src="https://img.shields.io/static/v1.svg?label=Contributions&message=Welcome&color=2267a0"\n     height="20"\n     alt="Contributions Welcome">\n  </a>\n  <h3>OpenRPC provides classes to rapidly develop an\n  <a href="https://open-rpc.org">OpenRPC</a> server.</h3>\n</div>\n\n## Installation\n\nOpenRPC is on PyPI and can be installed with:\n\n```shell\npip install openrpc\n```\n\n## Usage\n\nThis library provides an `OpenRPCServer` class that can be used to\nquickly create an OpenRPCServer; it takes as an argument an `InfoObject`\nwhich needs at minimum a title and version.\n\n```python\nfrom openrpc.objects import InfoObject\nfrom openrpc.server import OpenRPCServer\n\nrpc = OpenRPCServer(InfoObject(title="Demo Server", version="1.0.0"))\n```\n\n### Register a function as an RPC Method\n\nTo register a method with the OpenRPCServer add the `@rpc.method`\ndecorator to a method.\n\n```python\n@rpc.method\ndef add(a: int, b: int) -> int:\n    return a + b\n```\n\n#### RPC Discover\n\nThe `rpc.discover` method is automatically generated. It relies heavily\non type hints.\n\n### Process JSON RPC Request\n\nOpenRPC is transport agnostic. To use it, pass JSON RPC requests to the\n`process_request` method.\n\n```python\nreq = """\n{\n  "id": 1,\n  "method": "add",\n  "params": {"a": 2, "b": 2},\n  "jsonrpc": "2.0"\n}\n"""\nrpc.process_request(req)  # \'{"id": 1, "result": 4, "jsonrpc": "2.0}\'\n```\n\n### Async Support (v1.2+)\n\nOpenRPC has async support:\n\n```python\nawait rpc.process_request_async(req)\n```\n\n### Pydantic Support\n\nFor data classes, it is strongly recommended you use Pydantic.\nOpenRPCServer will use Pydantic for JSON serialization/deserialization\nas well as generating schemas.\n\n## Example\n\n```python\nfrom flask import Flask, Response, request\nfrom openrpc.objects import InfoObject\nfrom openrpc.server import OpenRPCServer\nfrom pydantic import BaseModel\n\napp = Flask(__name__)\nrpc = OpenRPCServer(InfoObject(title="Demo Server", version="1.0.0"))\n\n\nclass Vector3(BaseModel):\n    x: float\n    y: float\n    z: float\n\n\n@rpc.method\ndef get_distance(a: Vector3, b: Vector3) -> Vector3:\n    return Vector3(\n        x=a.x - b.x,\n        y=a.y - b.y,\n        z=a.z - b.z,\n    )\n\n\n@rpc.method\ndef add(a: int, b: int) -> int:\n    return a + b\n\n\n@app.route("/api/v1/", methods=["POST"])\ndef process_rpc() -> str:\n    return rpc.process_request(request.data)\n\n\nif __name__ == "__main__":\n    app.run()\n```\n\nExample In\n\n```json\n[\n  {\n    "id": 1,\n    "method": "add",\n    "params": {"a": 1, "b": 3},\n    "jsonrpc": "2.0"\n  }, {\n    "id": 2,\n    "method": "add",\n    "params": [5, 7],\n    "jsonrpc": "2.0"\n  }, {\n    "id": 3,\n    "method": "add",\n    "params": [11, "thirteen"],\n    "jsonrpc": "2.0"\n  }\n]\n```\n\nExample Result Out\n\n```json\n[\n  {\n    "id": 1,\n    "result": 4,\n    "jsonrpc": "2.0"\n  }, {\n    "id": 2,\n    "result": 12,\n    "jsonrpc": "2.0"\n  }, {\n    "id": 3,\n    "error": {\n      "code": -32000,\n      "message": "TypeError: unsupported operand type(s) for +: \'int\' and \'str\'"\n    },\n    "jsonrpc": "2.0"\n  }\n]\n```\n\nExample RPC Discover\n\n```json\n{\n  "id": 1,\n  "result": {\n    "openrpc": "1.2.6",\n    "info": {\n      "title": "Demo Server",\n      "version": "1.0.0"\n    },\n    "methods": [\n      {\n        "name": "add",\n        "params": [\n          {\n            "name": "a",\n            "schema": {\n              "type": "number"\n            },\n            "required": true\n          },\n          {\n            "name": "b",\n            "schema": {\n              "type": "number"\n            },\n            "required": true\n          }\n        ],\n        "result": {\n          "name": "result",\n          "schema": {\n            "type": "number"\n          },\n          "required": true\n        }\n      }\n    ],\n    "components": {\n      "schemas": {}\n    }\n  },\n  "jsonrpc": "2.0"\n}\n```\n',
    'author': 'Matthew Burkard',
    'author_email': 'matthewjburkard@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/mburkard/openrpc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Provides OpenRPCServer class."""
import inspect
import logging
import re
from functools import partial
from typing import (
    Any,
    Callable,
    Optional,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from openrpc._rpc_server import RPCServer
from openrpc.objects import (
    ComponentsObject,
    ContentDescriptorObject,
    InfoObject,
    MethodObject,
    OpenRPCObject,
    SchemaObject,
)

__all__ = ("OpenRPCServer",)

T = Type[Callable]
log = logging.getLogger("openrpc")


class OpenRPCServer:
    """OpenRPC server to register methods with."""

    def __init__(self, info: InfoObject, server_error_code: int = -32000) -> None:
        # This class wraps RPCServer which registers and executes methods.
        self.server = RPCServer(server_error_code)
        self.info: InfoObject = info
        self.components: ComponentsObject = ComponentsObject(schemas={})
        self.server.method(self.discover, method=MethodObject(name="rpc.discover"))

    def method(
        self, *args: Union[T, tuple[T]], method: Optional[MethodObject] = None
    ) -> T:
        """Register a method with this OpenRPC server.

        Can be used as a plain decorator, eg:

        .. code-block::

            @method
            def my_func()

        Or additional method info can be provided with a MethodObject:

        .. code-block::

            @method(method=MethodObject(deprecated=True))
            def my_func()

        :param args: The method if this is used as a plain decorator.
        :param method: MethodObject if addition information is required.
        :return: None
        """
        if args:
            func = args[0]
            method = MethodObject()
            return self.server.method(func, method)
        return partial(self.server.method, method=method)

    def process_request(self, data: Union[bytes, str]) -> Optional[str]:
        """Process a JSON-RPC2 request and get the response.

        :param data: A JSON-RPC2 request.
        :return: A valid JSON-RPC2 response.
        """
        log.debug("Processing request: %s", data)
        resp = self.server.process(data)
        log.debug("Responding : %s", resp)
        return resp

    async def process_request_async(self, data: Union[bytes, str]) -> Optional[str]:
        """Process a JSON-RPC2 request and get the response.

        If the method called by the request is async it will be awaited.

        :param data: A JSON-RPC2 request.
        :return: A valid JSON-RPC2 response.
        """
        log.debug("Processing request: %s", data)
        resp = await self.server.process_async(data)
        log.debug("Responding : %s", resp)
        return resp

    def discover(self) -> dict[str, Any]:
        """The OpenRPC discover method."""
        for name, rpc_method in self.server.methods.items():
            if name == "rpc.discover":
                continue
            method = rpc_method.method
            method.params = method.params or self._get_params(rpc_method.fun)
            method.result = method.result or self._get_result(rpc_method.fun)
        return OpenRPCObject(
            openrpc="1.2.6",
            info=self.info,
            methods=[
                it.method
                for it in self.server.methods.values()
                if it.method.name != "rpc.discover"
            ],
            components=self.components,
        ).dict(by_alias=True, exclude_unset=True)

    def _get_params(self, fun: Callable) -> list[ContentDescriptorObject]:
        # noinspection PyUnresolvedReferences,PyProtectedMember
        has_default = {
            k
            for k, v in inspect.signature(fun).parameters.items()
            if v.default != inspect._empty
        }
        return [
            ContentDescriptorObject(
                name=name,
                schema=self._get_schema(annotation),
                required=name not in has_default and self._is_required(annotation),
            )
            for name, annotation in get_type_hints(fun).items()
            if name != "return"
        ]

    def _get_result(self, fun: Callable) -> ContentDescriptorObject:
        return ContentDescriptorObject(
            name="result",
            schema=self._get_schema(get_type_hints(fun)["return"]),
            required=self._is_required(get_type_hints(fun)["return"]),
        )

    def _get_schema(self, annotation: Type) -> SchemaObject:
        if annotation == Any:
            return SchemaObject()
        if get_origin(annotation) == Union:
            return SchemaObject(
                anyOf=[self._get_schema(a) for a in get_args(annotation)]
            )

        schema_type = self._py_to_schema_type(annotation)

        if schema_type == "object":
            try:
                name = annotation.__name__
            except AttributeError:
                name = None
            if "schema" in dir(annotation):
                # noinspection PyUnresolvedReferences
                schema = SchemaObject(**annotation.schema())
                schema.title = schema.title or name
                for k, v in (schema.definitions or {}).items():
                    if k not in self.components.schemas:
                        self.components.schemas[k] = v
                # pydantic creates definitions, move them to components.
                for prop in schema.properties.values():
                    if prop.ref:
                        prop.ref = re.sub(
                            r"^#/definitions", "#/components/schemas", prop.ref
                        )
                del schema.definitions
            elif get_origin(annotation) == dict:
                schema = SchemaObject()
                schema.type = schema_type
                schema.additional_properties = True
                return schema
            else:
                schema = SchemaObject()
                schema.type = schema_type
                schema.properties = {
                    k: self._get_schema(v)
                    for k, v in get_type_hints(annotation).items()
                    if k != "return"
                }
            if schema not in self.components.schemas.values():
                self.components.schemas[name] = schema
            return SchemaObject(**{"$ref": f"#/components/schemas/{name}"})

        if schema_type == "array":
            schema = SchemaObject()
            schema.type = schema_type
            if args := get_args(annotation):
                schema.items = SchemaObject(**self._get_properties(args[0]))
            return schema

        schema = SchemaObject()
        schema.type = schema_type
        return schema

    def _py_to_schema_type(self, annotation: Any) -> Union[str, list[str]]:
        py_to_schema = {
            None: "null",
            str: "string",
            int: "number",
            float: "number",
            bool: "boolean",
        }
        origin = get_origin(annotation)
        flat_collections = [list, set, tuple]
        if origin in flat_collections or annotation in flat_collections:
            return "array"
        if dict in [origin, annotation]:
            return "object"
        if Union in [origin, annotation]:
            return self._py_to_schema_type(get_args(annotation)[0])
        if args := get_args(annotation):
            return [
                self._py_to_schema_type(arg)
                if "__name__" in dir(arg) and arg.__name__ != "NoneType"
                else "null"
                for arg in args
            ]
        if type(None) is annotation:
            return "null"
        return py_to_schema.get(annotation) or "object"

    def _get_properties(self, annotation: Type) -> dict[str, Any]:
        schema = self._get_schema(annotation)
        properties = {}
        if isinstance(schema, list):
            types = [arg.ref if arg.ref else arg.type for arg in schema]
            types = list(dict.fromkeys(types))
            if len(types) > 1:
                # noinspection PyTypedDict
                properties["type"] = types
                return properties
            schema = schema[0]
        if schema.ref:
            properties["$ref"] = schema.ref
        else:
            properties["type"] = schema.type
        return properties

    @staticmethod
    def _is_required(annotation: Any) -> bool:
        def _get_name(arg: Any) -> str:
            try:
                return arg.__name__
            except AttributeError:
                return ""

        return "NoneType" not in [_get_name(a) for a in get_args(annotation)]
